% AVL trees
%
% https://two-wrongs.com/purely-functional-avl-trees-in-common-lisp.html
%
%   node(KEY?, VAL?, TREE^)
%       Create a new TREE with a single entry.
%
%   height(TREE?, HEIGHT^)
%       Computes the height of TREE.
%
%   insert(KEY?, VAL?, TREE?, RTREE^)
%       Insert new entry into TREE, assigning the new tree to RTREE.
%
%   delete(KEY?, TREE?, [VAL^, ]RTREE^)
%       Delete entry from TREE and assign new tree to RTREE, optionally
%       assigning old existing value (or {}) to VAL.
%
%   replace(KEY?, VAL?, TREE?, [OLDVAL^, ]RTREE^[, DONE^])
%       Replace existing entry in TREE with new value, optionally assigning
%       the old value (or {}) to OLDVAL. RTREE holds the final result tree.
%       DONE is assigned the empty list when OLDVAL has been bound.
%
%   lookup(KEY?, TREE?, VAL^)
%       Assigns the value for the entry with the given KEY in TREE to VAL,
%       assigns {} if no such entry exists.

-exports([node/3, height/2, insert/4, delete/3, delete/4]).
-exports([replace/4, replace/5, replace/6, lookup/3]).

height({}, H) :- H := 0.
height({_, _, L, R}, H) :-
    height(L, H1), height(R, H2), H is 1 + max(H1, H2).

balance({}, B) :- B := 0.
balance({_, _, L, R}, B) :-
    height(L, HL), height(R, HR), B is HR - HL.

rotate_l({K, V, L, {KR, VR, LR, RR}}, T) :-
    node(K, V, L, LR, T1),
    node(KR, VR, T1, RR, T).

rotate_r({K, V, {KL, VL, LL, RL}, R}, T) :-
    node(K, V, RL, R, T1),
    node(KL, VL, LL, T1, T).

node(K, V, T) :- node(K, V, {}, {}, T).

node(K, V, L, R, T) :-
    balance({K, V, L, R}, B),
    node2(B, K, V, L, R, T).

node2(-2, K, V, L, R, T) :-
    balance(L, BL),
    node_il(BL, K, V, L, R, T).
node2(2, K, V, L, R, T) :-
    balance(R, BR),
    node_ir(BR, K, V, L, R, T).
node2(_, K, V, L, R, T) :- 
    otherwise | T := {K, V, L, R}.

node_il(-1, K, V, L, R, T) :- 
    rotate_r({K, V, L, R}, T).
node_il(1,  K, V, L, R, T) :- 
    rotate_l(L, L1),
    node(K, V, L1, R, T).

node_ir(-1, K, V, L, R, T) :- 
    rotate_r(R, R1),
    node(K, V, L, R1, T).
node_ir(1, K, V, L, R, T) :- 
    rotate_l({K, V, L, R}, T).

insert(K, V, {}, T) :- node(K, V, T).
insert(K, V, {KT, VT, L, R}, T) :-
    K @< KT |
    insert(K, V, L, TL),
    node(KT, VT, TL, R, T).
insert(K, V, {KT, VT, L, R}, T) :-
    otherwise |
    insert(K, V, R, TR),
    node(KT, VT, L, TR, T).

delete(K, T1, T) :- delete(K, T1, _, T).

delete(_, {}, V, T) :- T := {}, V := {}.
delete(K, {KT, VT, L, R}, V, T) :-
    K @< KT |
    delete(K, L, V, TL),
    node(KT, VT, TL, R, T).
delete(K, {KT, VT, L, R}, V, T) :-
    K @> KT |
    delete(K, R, V, TR),
    node(KT, VT, L, TR, T).
delete(_, {_, V1, L, R}, V, T) :-
    otherwise | 
    V := V1, delete2(L, R, T).

delete2({}, R, T) :- T := R.
delete2(L, {}, T) :- T := L.
delete2({_, _, LL, _}, R, T) :-
    otherwise | 
    min_node(R, MR),
    delete3(MR, LL, R, T).

delete3({K, V, _, _}, RL, RR, T) :-
    delete(K, RR, DR),
    node(K, V, RL, DR, T).

min_node({K, V, {}, R}, T) :- T := {K, V, {}, R}.
min_node({_, _, L, _}, T) :- otherwise | min_node(L, T).

replace(K, V, T1, T) :- replace1(T1, K, V, _, T, _).

replace(K, V, T1, V0, T) :- replace1(T1, K, V, V0, T, _).

replace(K, V, T1, V0, T, OK) :- replace1(T1, K, V, V0, T, OK).

replace1({}, _, _, V0, T, OK) :- 
    T := {}, V0 := {}, OK := [].
replace1({K1, V1, L, R}, K, V, V0, T, OK) :-
    K1 @< K | 
    replace1(L, K, V, V0, L1, OK),
    T := {K1, V1, L1, R}.
replace1({K1, V1, L, R}, K, V, V0, T, OK) :-
    K1 @> K | 
    replace1(R, K, V, V0, R1, OK),
    T := {K1, V1, L, R1}.
replace1({K, V1, L, R}, _, V, V0, T, OK) :- 
    otherwise | 
    assign(V0, V1, OK), T := {K, V, L, R}.

lookup(_, {}, X) :- X := {}.
lookup(K1, {K, _, L, _}, X) :-
    K1 @< K | lookup(K1, L, X).
lookup(K1, {K, _, _, R}, X) :-
    K1 @> K | lookup(K1, R, X).
lookup(_, {_, V, _, _}, X) :- X := V.
