% Strand compiler

-exports([init_state/5, compile_forms/3]).

compile_forms(FORMS, ST, DONE) :-
    declarations(FORMS, CLAUSES, ST, S2),
    group(CLAUSES, PDEFS, S2, S3),
    check_dups(PDEFS, S3, S4),
    check_exports(PDEFS, S4, S5),
    output_file(S5, ONAME),
    open_file(ONAME, w, OUT),
    compile_module(PDEFS, S5, PCODE),
    write_module(PCODE, [], OUT, DONE).

% state: {INAME, ONAME, EXPORTS, VARTAG, ERRSTREAM}
init_state(IN, ON, S, VT, E) :- 
    VT is random(100000), 
    S := {IN, ON, [], VT, E}.

% state accessors

output_file({_, ON, _, _, _}, R) :- R := ON.
exports({_, _, EXP, _, _}, R) :- R := EXP.
var_tag({_, _, _, VT, _}, R) :- R := VT.
errors({_, _, _, _, E}, R) :- R := E.

add_export(NA, {IN, ON, XL, VT, E}, S) :- 
    S := {IN, ON, [NA|XL], VT, E}.

close_errors({_, _, _, _, E}) :- E := [].

% declaration processing

declarations([], CS, S1, S) :- CS := [], S := S1.
declarations([-DECL|R], CS, S1, S) :-
    declaration(DECL, S1, S2),
    declarations(R, CS, S2, S).
declarations([C|R], CS, S1, S) :- 
    otherwise | 
    CS := [C|CS2], 
    declarations(R, CS2, S1, S).

declaration(exports(LST), S1, S) :- 
    add_exports(LST, S1, S).
declaration(DECL, S1, S) :-
    otherwise |
    compile_error('invalid declaration: ~q~n', [DECL], S1, S).

add_exports([], S1, S) :- S := S1.
add_exports([N/A|R], S1, S) :-
    add_export(N/A, S1, S2),
    add_exports(R, S2, S).
add_exports([X|_], S1, S) :-
    otherwise |
    compile_error('invalid export specification: ~q~n', [X], S1, S).

% various checks

check_dups([], S1, S) :- S := S1.
check_dups([{N, A, _}|PDEFS], S1, S) :-
    check_dups(PDEFS, N, A, S1, S2),
    check_dups(PDEFS, S2, S).

check_dups([], _, _, S1, S) :- S := S1.
check_dups([{N, A, _}|_], N, A, S1, S) :-
    compile_error('discontiguous process definition: ~a/~d\n', [N, A], S1, S).
check_dups([_|PDEFS], N, A, S1, S) :-
    otherwise | 
    check_dups(PDEFS, N, A, S1, S).

check_exports(PDEFS, S1, S) :-
    exports(S1, XL),
    check_exports1(XL, PDEFS, S1, S).

check_exports1([], _, S1, S) :- S := S1.
check_exports1([N/A|XL], PDEFS, S1, S) :-
    check_exports2(PDEFS, N, A, S1, S2),
    check_exports1(XL, PDEFS, S2, S).

check_exports2([], N, A, S1, S) :-
    compile_error('exported process is not defined: ~a/~d\n', [N, A], S1, S).
check_exports2([{N, A, _}|_], N, A, S1, S) :-
    S := S1.
check_exports2([_|PDEFS], N, A, S1, S) :-
    otherwise | 
    check_exports2(PDEFS, N, A, S1, S).

check_definition(NA, S1, S) :-
    internal(NA, K),
    check_definition2(K, NA, S1, S).

check_definition2(user, _, S1, S) :- S := S1.
check_definition2(none, NA, S1, S) :- 
    special(NA, F),
    check_definition2(F, NA, S1, S).
check_definition2(K, N/A, S1, S) :-
    K =\= none, K =\= user |
    compile_error('redefinition of primitive process ~s/~d\n', [N, A], S1, S).

check_singletons([], _, _, S1, S) :- S := S1.
check_singletons([I|SS], VS, T, S1, S) :-
    check_singletons2(VS, I, T, S1, S2),
    check_singletons(SS, VS, T, S2, S).

check_singletons2([], _, _, S1, S) :- S := S1.
check_singletons2([{I, N}|_], I, T, S1, S) :-
    N =\= '_' |
    head(T, DN, A, S1, S2),
    compile_warning('warning: singleton variable in definition of ~a/~d: ~a\n', 
        [DN, A, N], S2, S).
check_singletons2([_|VS], I, T, S1, S) :-
    otherwise | check_singletons2(VS, I, T, S1, S).

% write module to file
write_module([], _, _, DONE) :- DONE := [].
write_module([CODE|R], [], OUT, DONE) :-
    write_bytes(OUT, CODE, DONE1), 
    write_module(R, DONE1, OUT, DONE).

% flatten collected code fragments into byte list

flatten_code(CODE, CLIST) :- 
    flatten_code(CODE, CLIST, []).

flatten_code([], C, T) :- C := T.
flatten_code([X|R], C, T) :-
    flatten_code1(X, C, T1),
    flatten_code(R, T1, T).

flatten_code1(X, C, T) :- 
    integer(X) | integer_to_list(X, L), list:append(L, T, C).
flatten_code1(X, C, T) :-
    string(X) | string_to_byte_list(X, C, T).
flatten_code1(X, _, _) :-
    otherwise | error(internal_compiler_error('bad code expression', X)).

% toplevel

compile_module(PDEFS, S, SCODE) :-
    gen_mprefix(PDEFS, PCODE),
    SCODE := [PCODE|STAIL],
    compile_module2(PDEFS, 0, PDEFS, S, STAIL).

compile_module2([], _, _, S, SCODE) :- 
    SCODE := [],
    close_errors(S).
compile_module2([{N, A, G}|R], I, PDEFS, S, SCODE) :-
    compile_group(G, N, A, I, PDEFS, S, S2, SCODE, STAIL),
    I2 is I + 1,
    compile_module2(R, I2, PDEFS, S2, STAIL).
  
gen_mprefix(PDEFS, SCODE) :-
    length(PDEFS, N),
    N2 is N * 2,
    CODE1 := ['T', N2, ':'],
    flatten_code(CODE1, SCODE).

% compile group of process definitions 

compile_group(G, N, A, I, PDEFS, S1, S, SCODE, STAIL) :-
    exports(S1, XL),
    var_tag(S1, VTAG),
    check_definition(N/A, S1, S2),
    gen_entry(N, A, CODE, CODE1),
    term:classify_arg(G, VTAG, TL, LL, AL, OL),   
    compile_pdef(G, PDEFS, S2, S, [TL, LL, AL, OL], CODE1),
    flatten_code(CODE, STR),
    length(STR, CLEN),
    LEN is CLEN - 1,    % subtract trailing newline
    compile_prefix(XL, N, A, I, LEN, PREF),
    SCODE := [PREF, STR|STAIL].

gen_entry(N, A, CODE, TAIL) :-
    byte_length(N, NLEN), 
    integer_to_list(A, AL), 
    length(AL, ALEN),
    LEN is NLEN + ALEN + 1,
    CODE := ['D', LEN, ':', N, '/', A, 'I', A, ':A'|TAIL].

compile_prefix([], _, _, I, CLEN, PREF) :-
    CODE := ['I', I, ':$', CLEN, ':'],      % unexported
    flatten_code(CODE, PREF).
compile_prefix([N/A|_], N, A, _, CLEN, PREF) :-
    byte_length(N, NLEN),       % exported
    CODE := ['T2:$', NLEN, ':', N, 'I', A , ':$', CLEN, ':'],
    flatten_code(CODE, PREF).
compile_prefix([_|R], N, A, I, CLEN, PREF) :-
    otherwise | compile_prefix(R, N, A, I, CLEN, PREF).

compile_pdef(G, PDEFS, S1, S, [[], [], [], _], CODE) :-
    compile_pdef1(G, PDEFS, '[', S1, S, CODE, TAIL),
    TAIL := [']\n'].
compile_pdef(_, PDEFS, S1, S, SL, CODE) :-
    otherwise | 
    CODE := ['[SS'|TAIL1],
    compile_switch_pdef(SL, ['ST', 'SL', 'SA', 'SO'], '\'', PDEFS, S1, S, TAIL1).

compile_pdef1([], _, _, S1, S, CODE, TAIL) :- 
    CODE := TAIL, S := S1.
compile_pdef1([T|G], PDEFS, TRY, S1, S, CODE, TAIL) :-
    CODE := [TRY|TAIL1],
    var_tag(S1, VTAG),
    classify_variables(T, VTAG, VS, SS, S1, S2),    
    compile_term(T, VS, SS, PDEFS, S2, S3, TAIL1, TAIL2),
    compile_pdef1(G, PDEFS, '\'', S3, S, TAIL2, TAIL).

compile_switch_pdef([], _, _, _, S1, S, CODE) :- 
    CODE := [']\n'], S := S1.
compile_switch_pdef([[]], ['SO'], _, _, S1, S, CODE) :-
    % case of no "other" set - patch up `switch` op anyway
    CODE := ['SO]\n'], S := S1.
compile_switch_pdef([[]|RG], [_|SR], TRY, PDEFS, S1, S, CODE) :-
    compile_switch_pdef(RG, SR, TRY, PDEFS, S1, S, CODE).
compile_switch_pdef([G|RG], [SOP|SR], TRY, PDEFS, S1, S, CODE) :-
    G =\= [] |
    CODE := [SOP|TAIL1],
    compile_pdef1(G, PDEFS, TRY, S1, S2, TAIL1, TAIL2),
    compile_switch_pdef(RG, SR, '\'', PDEFS, S2, S, TAIL2).

% separate singleton from non-singleton variables

classify_variables(T, VTAG, VS, SS, S1, S) :-
    term:vars(T, VTAG, VS1),
    var_indices(VS1, VS2),
    sort:sort(VS2, VS),
    list:unique(VS, VS2, SS),
    check_singletons(SS, VS1, T, S1, S).

var_indices([], IS) :- IS := [].
var_indices([{I, _}|VS], IS) :- IS := [I|IS2], var_indices(VS, IS2).

% compile single process definition

compile_term((H :- T), VS, SS, PDEFS, S1, S, CODE, TAIL) :-
    compile_term1(H, T, VS, SS, PDEFS, S1, S, CODE, TAIL).
compile_term(H, VS, SS, PDEFS, S1, S, CODE, TAIL) :-
    compile_term1(H, true, VS, SS, PDEFS, S1, S, CODE, TAIL).

compile_term1(H, T, VS, SS, PDEFS, S1, S, CODE, TAIL) :-
    term:args(H, ARGS),
    compile_head(ARGS, VS, SS, VSS, S1, S2, CODE, TAIL1),
    compile_tail(T, VS, SS, VSS, S2, S, PDEFS, TAIL1, TAIL).

compile_head(ARGS, VS, SS, VSS, S1, S, CODE, TAIL) :-
    length(VS, ESIZE),
    compile_head1(ESIZE, ARGS, SS, VSS, S1, S, CODE, TAIL).

compile_head1(0, ARGS, SS, VSS, S1, S, CODE, TAIL) :-
    compile_head_arguments(ARGS, 0, SS, [], VSS, S1, S, CODE, TAIL).
compile_head1(N, ARGS, SS, VSS, S1, S, CODE, TAIL) :-
    N > 0 | 
    CODE := ['I', N, ':V'|TAIL1],
    compile_head_arguments(ARGS, 0, SS, [], VSS, S1, S, TAIL1, TAIL).

compile_head_arguments([], _, _, VSS, VSS2, S1, S, CODE, TAIL) :- 
    CODE := TAIL, VSS2 := VSS, S := S1.
compile_head_arguments([X|R], I, SS, VSS, VSS2, S1, S, CODE, TAIL) :-
    CODE := ['I', I, ':G'|TAIL1],
    var_tag(S1, VTAG), 
    compile_match(X, VTAG, SS, VSS, VSS1, S1, S2, TAIL1, TAIL2),
    I2 is I + 1,
    compile_head_arguments(R, I2, SS, VSS1, VSS2, S2, S, TAIL2, 
        TAIL).

% compile head matching

compile_match('$VAR'(TAG, VI, _), TAG, SS, VSS, VSS2, S1, S, CODE, TAIL) :-
    S := S1,
    compile_match_var(SS, VI, VSS, VSS2, CODE, TAIL).
compile_match(T, VTAG, SS, VSS, VSS2, S1, S, CODE, TAIL) :-
    tuple(T) |
    length(T, TLEN),
    CODE := ['I', TLEN, ':N'|TAIL1],
    tuple_to_list(T, TL, []),
    compile_tuple_match(TL, VTAG, SS, VSS, VSS2, S1, S, TAIL1, TAIL).
compile_match([], _, _, VSS, VSS2, S1, S, CODE, TAIL) :-
    CODE := ['%'|TAIL], VSS2 := VSS, S := S1.
compile_match(X, _, _, VSS, VSS2, S1, S, CODE, TAIL) :-
    integer(X) | 
    CODE := ['I', X, ':U'|TAIL], VSS2 := VSS, S := S1.
compile_match(X, _, _, VSS, VSS2, S1, S, CODE, TAIL) :-
    string(X) | 
    byte_length(X, LEN),
    CODE := ['$', LEN, ':', X, '='|TAIL], VSS2 := VSS, S := S1.
compile_match([X|Y], VTAG, SS, VSS, VSS2, S1, S, CODE, TAIL) :-
    CODE := ['L'|TAIL1],
    compile_match(X, VTAG, SS, VSS, VSS1, S1, S2, TAIL1, TAIL2),
    compile_match(Y, VTAG, SS, VSS1, VSS2, S2, S, TAIL2, TAIL).
compile_match(X, _, _, _, VSS2, S1, S, CODE, TAIL) :-
    otherwise | 
    VSS2 := [], CODE := TAIL,
    compile_error('invalid match value: ~q~n', [X], S1, S).

compile_match_var([], VI, VSS, VSS2, CODE, TAIL) :-
    compile_match_var1(VSS, VI, VSS, VSS2, CODE, TAIL).
compile_match_var([VI|_], VI, VSS, VSS2, CODE, TAIL) :-
    CODE := ['O'|TAIL], VSS2 := VSS. % singleton
compile_match_var([_|R], VI, VSS, VSS2, CODE, TAIL) :-
    otherwise | 
    compile_match_var(R, VI, VSS, VSS2, CODE, TAIL).

compile_match_var1([], VI, VSS, VSS2, CODE, TAIL) :-
    CODE := ['I', VI, ':P'|TAIL], VSS2 := [VI|VSS]. % first occurrence
compile_match_var1([VI|_], VI, VSS, VSS2, CODE, TAIL) :-
    CODE := ['I', VI, ':M'|TAIL], VSS2 := VSS.  % subseq. occurrence
compile_match_var1([_|R], VI, VSS, VSS2, CODE, TAIL) :-
    otherwise | 
    compile_match_var1(R, VI, VSS, VSS2, CODE, TAIL).

compile_tuple_match([], _, _, VSS, VSS2, S1, S, CODE, TAIL) :-
    VSS2 := VSS, CODE := TAIL, S := S1.
compile_tuple_match([X|R], VTAG, SS, VSS, VSS2, S1, S, CODE, TAIL) :-
    compile_match(X, VTAG, SS, VSS, VSS1, S1, S2, CODE, TAIL1),
    compile_tuple_match(R, VTAG, SS, VSS1, VSS2, S2, S, TAIL1, TAIL).

compile_tail((G | B), VS, SS, VSS, S1, S, PDEFS, CODE, TAIL) :-
    list:difference(VS, VSS, SSG),
    compile_guards(G, SSG, S1, S2, CODE, TAIL1),
    compile_body(B, VS, SS, VSS, S2, S, PDEFS, TAIL1, TAIL).
compile_tail(B, VS, SS, VSS, S1, S, PDEFS, CODE, TAIL) :-
    otherwise | 
    compile_body(B, VS, SS, VSS, S1, S, PDEFS, CODE, TAIL).

% compile guards

compile_guards((G1, G2), SS, S1, S, CODE, TAIL) :-
    compile_guard(G1, SS, S1, S2, CODE, TAIL1),
    compile_guards(G2, SS, S2, S, TAIL1, TAIL).
compile_guards(G, SS, S1, S, CODE, TAIL) :-
    otherwise | 
    compile_guard(G, SS, S1, S, CODE, TAIL).

compile_guard(otherwise, _, S1, S, CODE, TAIL) :- 
    CODE := TAIL, S := S1.
compile_guard(X, SS, S1, S, CODE, TAIL) :-
    tuple(X) |
    tuple_to_list(X, XL, []),
    list:pair(XL, N, ARGS),
    length(ARGS, A),
    guard(N/A, K),
    compile_guard1(K, X, ARGS, SS, S1, S, CODE, TAIL).
compile_guard(X, _, S1, S, CODE, TAIL) :-
    otherwise | 
    CODE := TAIL, 
    compile_error('invalid guard expression: ~q\n', [X], S1, S).

compile_guard1(none, X, _, _, S1, S, CODE, TAIL) :-
    CODE := TAIL,
    compile_error('invalid guard: ~q~n', [X], S1, S).
compile_guard1(K/n, X, ARGS, SS, S1, S, CODE, TAIL) :- 
    var_tag(S1, VT),
    compile_expressions(ARGS, VT, SS, X, S1, S, CODE, [K|TAIL]).
compile_guard1(K, _, ARGS, SS, S1, S, CODE, TAIL) :- 
    otherwise |
    compile_values(ARGS, SS, S1, S, CODE, [K|TAIL]).

% compile term arguments

compile_values([], _, S1, S, CODE, TAIL) :- 
    CODE := TAIL, S := S1.
compile_values([X|R], SS, S1, S, CODE, TAIL) :-
    var_tag(S1, VTAG),
    compile_value(X, VTAG, SS, S1, S2, CODE, TAIL1),
    compile_values(R, SS, S2, S, TAIL1, TAIL).

compile_value('$VAR'(TAG, I, _), TAG, SS, S1, S, CODE, TAIL) :-
    S := S1,
    compile_value_var(SS, I, CODE, TAIL).
compile_value({}, _, _, S1, S, CODE, TAIL) :- 
    CODE := ['I0:B'|TAIL], S := S1.
compile_value(T, _, SS, S1, S, CODE, TAIL) :-
    tuple(T) | 
    tuple_to_list(T, TL, []), 
    list:reverse(TL, TL2),
    length(T, TLEN),
    compile_values(TL2, SS, S1, S, CODE, ['I', TLEN, ':B'|TAIL]).
compile_value(X, _, _, S1, S, CODE, TAIL) :-
    integer(X) | 
    CODE := ['z', X, ':'|TAIL], S := S1.
compile_value([], _, _, S1, S, CODE, TAIL) :- 
    CODE := [','|TAIL], S := S1.
compile_value(X, _, _, S1, S, CODE, TAIL) :-
    string(X) | 
    byte_length(X, XLEN),
    CODE := ['$', XLEN, ':', X|TAIL], S := S1.
compile_value([X|Y], VTAG, SS, S1, S, CODE, TAIL) :-
    compile_value(X, VTAG, SS, S1, S2, CODE, TAIL1),
    compile_value(Y, VTAG, SS, S2, S, TAIL1, ['.'|TAIL]).
compile_value(X, _, _, S1, S, CODE, TAIL) :-
    otherwise | 
    CODE := TAIL,
    compile_error('invalid value: ~q~n', [X], S1, S).

compile_value_var([], I, CODE, TAIL) :- 
    CODE := ['I', I, ':E'|TAIL].    % non-singleton
compile_value_var([I|_], I, CODE, TAIL) :- 
    CODE := ['?'|TAIL].     % singleton
compile_value_var([_|R], I, CODE, TAIL) :-
    otherwise | 
    compile_value_var(R, I, CODE, TAIL).

% compile body goals

compile_body(B, VS, SS, VSS, S1, S, PDEFS, CODE, TAIL) :-
    list:difference(VS, SS, VS1),
    assign_vars(VS1, VSS, CODE, TAIL1),
    reorder_body(B, FB),
    compile_goals(FB, SS, S1, S, PDEFS, TAIL1, TAIL).

reorder_body(B, BR) :-
    term:flatten_body(B, B1, BTL),
    reorder_body(B1, BTL, BR).

% move first goal to end, meeting user's expectations that goals are
% scheduled roughly in textual order.
reorder_body([G1|B], TL, BR) :-
    BR := B, TL := [G1].

assign_vars([], _, CODE, TAIL) :- CODE := TAIL.
assign_vars([VI|R], VSS, CODE, TAIL) :-
    assign_vars1(VSS, VI, CODE, TAIL1),
    assign_vars(R, VSS, TAIL1, TAIL).

assign_vars1([], VI, CODE, TAIL) :- CODE := ['?I', VI, ':P'|TAIL].
assign_vars1([VI|_], VI, CODE, TAIL) :- CODE := TAIL.
assign_vars1([_|R], VI, CODE, TAIL) :-
    otherwise | assign_vars1(R, VI, CODE, TAIL).

compile_goals([true], _, S1, S, _, CODE, TAIL) :-
    CODE := ['X'|TAIL], S := S1.
compile_goals([T], SS, S1, S, PDEFS, CODE, TAIL) :-
    T =\= true |
    CODE := [';'|TAIL1],
    compile_goal(T, SS, S1, S, PDEFS, TAIL1, TAIL).
compile_goals([T|B], SS, S1, S, PDEFS, CODE, TAIL) :-
    T =\= true, B =\= [] |
    CODE := ['('|TAIL1],
    compile_goal(T, SS, S1, S2, PDEFS, TAIL1, [')'|TAIL2]),
    compile_goals(B, SS, S2, S, PDEFS, TAIL2, TAIL).

compile_goal(true, _, S1, S, _, CODE, TAIL) :- 
    CODE := ['X'|TAIL], S := S1.
compile_goal(X := Y, SS, S1, S, _, CODE, TAIL) :-
    var_tag(S1, VTAG), 
    assignment_target(X, VTAG, VI, S1, S2),
    non_circular_assignment(Y, VI, VTAG, S2, S3),
    compile_values([X, Y], SS, S3, S, CODE, ['!X'|TAIL]).
compile_goal(X is Y, SS, S1, S, _, CODE, TAIL) :-
    var_tag(S1, VTAG), 
    compile_value(X, VTAG, SS, S1, S2, CODE, TAIL1),
    compile_expression(Y, SS, S2, S, TAIL1, ['#!X'|TAIL]).
compile_goal(run(T, ST, C), SS, S1, S, PDEFS, CODE, TAIL) :-
    functor(T, N, A, S1, S2),
    find_pdef(PDEFS, N, A, IX, S2, S3),
    term:args(T, ARGS),
    list_to_tuple(ARGS, ARGV),
    CODE := ['Hz', IX, ':'|TAIL1],
    % $run/5
    compile_values([ARGV, ST, C], SS, S3, S, TAIL1, ['K73:'|TAIL]).
compile_goal(T@NODE, SS, S1, S, PDEFS, CODE, TAIL) :-
    compile_remote_call(T, NODE, SS, S1, S, PDEFS, CODE, TAIL).
compile_goal(MN:CALL, _, S1, S, _, CODE, TAIL) :-
    string(CALL) |
    byte_length(MN, MLEN),
    byte_length(CALL, NLEN),
    S := S1,
    CODE := ['C', MLEN, ':', MN, NLEN, ':', CALL, 0, ':'|TAIL].
compile_goal(MN:CALL, SS, S1, S, _, CODE, TAIL) :-
    tuple(CALL) |
    tuple_to_list(CALL, TL, []),
    list:pair(TL, N, ARGS),
    byte_length(MN, MLEN),
    byte_length(N, NLEN),
    length(ARGS, A),
    compile_values(ARGS, SS, S1, S, CODE, ['C', MLEN, ':', MN, NLEN, 
        ':', N, A, ':'|TAIL]).
compile_goal(CALL, SS, S1, S, PDEFS, CODE, TAIL) :-
    string(CALL), CALL =\= true |
    compile_call(CALL, 0, [], SS, S1, S, PDEFS, CODE, TAIL).
compile_goal(CALL, SS, S1, S, PDEFS, CODE, TAIL) :-
    tuple(CALL) | 
    tuple_to_list(CALL, TL, []),
    list:pair(TL, N, ARGS),
    length(ARGS, A),
    compile_call(N, A, ARGS, SS, S1, S, PDEFS, CODE, TAIL).
compile_goal(X, _, S1, S, _, CODE, TAIL) :-
    otherwise | 
    CODE := TAIL,
    compile_error('invalid term: ~q~n', [X], S1, S).

compile_remote_call(MN:CALL, NODE, SS, S1, S, _, CODE, TAIL) :-
    functor(CALL, N, _, S1, S2),
    term:args(CALL, ARGS),
    list_to_tuple(ARGS, ARGV),
    % $call_remote/4
    compile_values([MN, N, ARGV, NODE], SS, S2, S, CODE, ['K8:'|TAIL]).
compile_remote_call(CALL, NODE, SS, S1, S, PDEFS, CODE, TAIL) :-
    functor(CALL, N, A, S1, S2),
    find_pdef(PDEFS, N, A, IX, S2, S3),
    term:args(CALL, ARGS),
    list_to_tuple(ARGS, ARGV),
    CODE := ['Hz', IX, ':'|TAIL1],
    % $call_remote/4
    compile_values([ARGV, NODE], SS, S3, S, TAIL1, ['K8:'|TAIL]).
compile_remote_call(CALL, _, _, S1, S, _, CODE, TAIL) :-
    otherwise | 
    CODE := TAIL,
    compile_error('invalid remote call: ~q~n', [CALL], S1, S).

compile_call(N, A, ARGS, SS, S1, S, PDEFS, CODE, TAIL) :-
    internal(N/A, K),
    compile_call1(K, N, A, ARGS, SS, S1, S, PDEFS, CODE, TAIL).

compile_call1(none, N, A, ARGS, SS, S1, S, PDEFS, CODE, TAIL) :-
    find_pdef(PDEFS, N, A, IX, S1, S2),
    compile_values(ARGS, SS, S2, S, CODE, ['J', IX, ':'|TAIL]).
compile_call1(K, _, _, ARGS, SS, S1, S, _, CODE, TAIL) :-
    compile_values(ARGS, SS, S1, S, CODE, ['K', K, ':X'|TAIL]).

% compile arithmetic expression

compile_expression(X, SS, S1, S, CODE, TAIL) :-
    var_tag(S1, VTAG),
    compile_expression1(X, VTAG, SS, X, S1, S, CODE, TAIL).

compile_expression1('$VAR'(TAG, I, _), TAG, SS, X, S1, S, CODE, TAIL) :-
    compile_expr_var(SS, I, X, S1, S, CODE, TAIL).
compile_expression1(X, _, _, _, S1, S, CODE, TAIL) :-
    integer(X) | 
    CODE := ['I', X, ':'|TAIL], S := S1.
compile_expression1(X, VTAG, SS, X0, S1, S, CODE, TAIL) :-
    string(X) | 
    expr(X/0, K), 
    compile_expression2(K, X, VTAG, SS, X0, [], S1, S, CODE, TAIL).
compile_expression1(X, VTAG, SS, X0, S1, S, CODE, TAIL) :-
    tuple(X) | 
    tuple_to_list(X, XL, []),
    list:pair(XL, N, ARGS),
    length(ARGS, A),
    expr(N/A, K),
    compile_expression2(K, X, VTAG, SS, X0, ARGS, S1, S, CODE, TAIL).
compile_expression1(X, _, _, X0, S1, S, CODE, TAIL) :-
    otherwise | 
    CODE := TAIL, 
    compile_error('invalid expression: ~q in: ~q~n', [X, X0], S1, S).

compile_expression2(none, X, _, _, X0, _, S1, S, CODE, TAIL) :-
    CODE := TAIL,
    compile_error('unknown expression operator ~q in: ~q~n', [X, X0], S1, S).
compile_expression2(K, _, VTAG, SS, X0, ARGS, S1, S, CODE, TAIL) :-
    otherwise |
    compile_expressions(ARGS, VTAG, SS, X0, S1, S, CODE, [K|TAIL]).

compile_expr_var([], I, _, S1, S, CODE, TAIL) :-
    CODE := ['I', I, ':E|'|TAIL], S := S1.
compile_expr_var([I|_], I, X, S1, S, CODE, TAIL) :-
    CODE := TAIL,
    compile_error('singleton variable in arithmetic expression: ~q~n', [X], S1, S).
compile_expr_var([J|R], I, X, S1, S, CODE, TAIL) :-
    J =\= I |
    compile_expr_var(R, I, X, S1, S, CODE, TAIL).

compile_expressions([], _, _, _, S1, S, CODE, TAIL) :- 
    CODE := TAIL, S := S1.
compile_expressions([X|R], VTAG, SS, X0, S1, S, CODE, TAIL) :-
    compile_expression1(X, VTAG, SS, X0, S1, S2, CODE, TAIL1),
    compile_expressions(R, VTAG, SS, X0, S2, S, TAIL1, TAIL).

% collect group of toplevel terms with same name + arity

group([], PDEFS, S1, S) :- 
    PDEFS := [], S := S1.
group([T|R], PDEFS, S1, S) :- 
    head(T, N, A, S1, S2), 
    PDEFS := [{N, A, [T|CS]}|PDEFS2],
    group1(R, N, A, CS, PDEFS2, S2, S).

group1([], _, _, CS, PDEFS, S1, S) :- 
    CS := [], PDEFS := [], S := S1.
group1([T|R], N, A, CS, PDEFS, S1, S) :-
    head(T, N2, A2, S1, S2), 
    group2(T, R, N, A, N2, A2, CS, PDEFS, S2, S).

group2(T, R, N, A, N, A, CS, PDEFS, S1, S) :-
    CS := [T|CS2],
    group1(R, N, A, CS2, PDEFS, S1, S).
group2(T, R, _, _, N2, A2, CS, PDEFS, S1, S) :-
    otherwise |
    CS := [],
    PDEFS := [{N2, A2, [T|CS2]}|PDEFS2],
    group1(R, N2, A2, CS2, PDEFS2, S1, S).

% support definitions

head((H :- _), N, A, S1, S) :- functor(H, N, A, S1, S).
head(T, N, A, S1, S) :- functor(T, N, A, S1, S).

functor(T, N, A, S1, S) :- 
    tuple(T) | 
    get_arg(1, T, N), length(T, LEN), A is LEN - 1, S := S1.
functor(T, N, A, S1, S) :-
    string(T) | N := T, A := 0, S := S1.
functor(T, N, A, S1, S) :-
    otherwise | 
    N := none, A := 0,
    compile_error('invalid process form: ~q\n', [T], S1, S).

find_pdef(PDEFS, N, A, IX, S1, S) :- 
    find_pdef1(PDEFS, N, A, 0, IX, S1, S).

find_pdef1([], N, A, _, IX, S1, S) :-
    IX := 0,
    compile_error('reference to undefined process: ~q\n', [N/A], S1, S).
find_pdef1([{N, A, _}|_], N, A, I, IX, S1, S) :- 
    IX := I, S := S1.
find_pdef1([_|R], N, A, I, IX, S1, S) :- 
    otherwise | 
    I2 is I + 1, 
    find_pdef1(R, N, A, I2, IX, S1, S).

assignment_target('$VAR'(TAG, I, _), TAG, VI, S1, S) :- 
    VI := I, S := S1.
assignment_target(X, _, VI, S1, S) :- 
    otherwise | 
    VI := 0,
    compile_error('invalid assignment target: ~q~n', [X], S1, S).

non_circular_assignment(X, I, VTAG, S1, S) :- 
    tuple(X) | 
    non_circular_assignment2(X, I, VTAG, S1, S).
non_circular_assignment([X|R], I, VTAG, S1, S) :- 
    non_circular_assignment(X, I, VTAG, S1, S2),
    non_circular_assignment(R, I, VTAG, S2, S).
non_circular_assignment(_, _, _, S1, S) :- 
    otherwise | S := S1.

non_circular_assignment2('$VAR'(TAG, I, NAME), I, TAG, S1, S) :-
    compile_error('circular assignment: ~a~n', [NAME], S1, S).
non_circular_assignment2(X, I, VTAG, S1, S) :- 
    otherwise |
    tuple_to_list(X, L, []), 
    non_circular_assignment(L, I, VTAG, S1, S).

% error reporting

compile_error(FMT, ARGS, {IN, ON, XL, VT, E}, S) :-
    E := [{0, FMT, ARGS, 1}|E2],
    S := {IN, ON, XL, VT, E2}.

compile_warning(FMT, ARGS, {IN, ON, XL, VT, E}, S) :-
    E := [{0, FMT, ARGS, 0}|E2],
    S := {IN, ON, XL, VT, E2}.

% guards + primitive processes
    
guard('=='/2, K) :- K := '='.
guard('=\\='/2, K) :- K := '`'.
guard('=:='/2, K) :- K := 'm'/n.
guard('\\=:='/2, K) :- K := 'o'/n.
guard('>'/2, K) :- K := '>'/n.
guard('<'/2, K) :- K := '<'/n.
guard('>='/2, K) :- K := 'g'/n.
guard('=<'/2, K) :- K := 'l'/n.
guard('@>'/2, K) :- K := 'a'.
guard('@<'/2, K) :- K := 'b'.
guard('@>='/2, K) :- K := 'h'.
guard('@=<'/2, K) :- K := 's'.
guard(known/1, K) :- K := 'k'.
guard(unknown/1, K) :- K := 'u'.
guard(data/1, K) :- K := 'd'.
guard(string/1, K) :- K := 'c'.
guard(tuple/1, K) :- K := 't'.
guard(list/1, K) :- K := 'j'.
guard(integer/1, K) :- K := 'i'.
guard(module/1, K) :- K := 'q'.
guard(_, K) :- otherwise | K := none.

expr('+'/2, K) :- K := '+'.
expr('-'/2, K) :- K := '-'.
expr('*'/2, K) :- K := '*'.
expr('/'/2, K) :- K := '/'.
expr('\\\\'/2, K) :- K := '\\'.
expr(abs/1, K) :- K := '&'.
expr('/\\'/2, K) :- K := '^'.
expr('\\/'/2, K) :- K := 'v'.
expr('><'/2, K) :- K := 'x'.
expr('<<'/2, K) :- K := '{'.
expr('>>'/2, K) :- K := '}'.
expr('\\'/1, K) :- K := '~'.
expr('-'/1, K) :- K := '_'.
expr(random/1, K) :- K := 'n'.
expr(isqrt/1, K) :- K := 'r'.
expr(clock/0, K) :- K := 'e'.
expr(max/2, K) :- K := 'W'.
expr(min/2, K) :- K := 'w'.
expr(_, K) :- otherwise | K := none.

internal('$reclaim'/2, K) :- K:= 0.
internal('$events'/1, K) :- K := 1.
internal('$send'/2, K) :- K := 2.
internal('$send'/3, K) :- K := 3.
internal('$node'/1, K) :- K := 4.
internal('$assign_remote'/3, K) :- K := 5.
internal('$get_exposed'/2, K) :- K := 6.
internal('$log'/1, K) :- K := 7.
internal('$call_remote'/4, K) :- K := 8.
internal('$call'/3, K) :- K := 9.
internal('$retrieve_module'/3, K) :- K := 10.
internal('$load_module'/2, K) :- K := 11.
internal(halt/1, K) :- K := 12.
internal(error/1, K) :- K := 13.
internal(write/1, K) :- K := 14.
internal(writeln/1, K) :- K := 15.
internal(string_to_list/3, K) :- K := 16.
internal(list_to_string/2, K) :- K := 17.
internal(make_tuple/2, K) :- K := 18.
internal(list_to_integer/2, K) :- K := 19.
internal(integer_to_list/2, K) :- K := 20.
internal(list_to_tuple/2, K) :- K := 21.
internal(tuple_to_list/3, K) :- K := 22.
internal(length/2, K) :- K := 23.
internal(run/2, K) :- K := 24.
internal(get_module/2, K) :- K := 25.
internal(assign/3, K) :- K := 26.
internal(put_arg/3, K) :- K := 27.
internal(put_arg/4, K) :- K := 28.
internal(get_arg/3, K) :- K := 29.
internal(get_arg/4, K) :- K := 30.
internal(open_file/3, K) :- K := 31.
internal(close_file/1, K) :- K := 32.
internal(read_bytes/4, K) :- K := 33.
internal(write_chars/2, K) :- K := 34.
internal(write/2, K) :- K := 35.
internal(merger/2, K) :- K := 36.
internal(list_to_integer/3, K) :- K := 37.
internal(integer_to_list/3, K) :- K := 38.
internal(byte_length/2, K) :- K := 39.
internal(write/3, K) :- K := 40.
internal(write_chars/3, K) :- K := 41.
internal(string_to_byte_list/3, K) :- K := 42.
internal(write_bytes/2, K) :- K := 43.
internal(write_bytes/3, K) :- K := 44.
internal(command_line/1, K) :- K := 45.
internal('$resolve_peer'/2, K) :- K := 46.
internal(utf_decode/3, K) :- K := 47.
internal('$register_peer'/2, K) :- K := 48.
internal('$node'/2, K) :- K := 49.
internal('$add_reference'/1, K) :- K := 50.
internal('$drop_reference'/1, K) :- K := 51.
internal(listen/2, K) :- K := 52.
internal('$forward'/2, K) :- K := 53.
internal('$deliver'/1, K) :- K := 54.
internal(strand_version/1, K) :- K := 55.
internal(randomize/1, K) :- K := 56.
internal('$statistics'/1, K) :- K := 57.
internal(close_file/2, K) :- K := 58.
internal(platform/2, K) :- K := 59.
internal(getenv/2, K) :- K := 60.
internal(shell/2, K) :- K := 61.
internal(open_pipe/2, K) :- K := 62.
internal('$execute'/7, K) :- K := 63.
internal('$read_forwarded'/2, K) :- K := 64.
internal(file_size/2, K) :- K := 65.
internal(file_modification_time/2, K) :- K := 66.
internal(string_to_integer/2, K) :- K := 67.
internal(string_to_integer/3, K) :- K := 68.
internal(run/4, K) :- K := 69.
internal(put_global/3, K) :- K := 70.
internal(get_global/2, K) :- K := 71.
internal(chdir/2, K) :- K := 72.
internal('$run'/5, K) :- K := 73.
internal(file_status/2, K) :- K := 74.
internal(time/1, K) :- K := 75.
internal(deref/2, K) :- K := 76.
internal(module_exports/2, K) :- K := 77.
internal(pledge/3, K) :- K := 100.
internal(unveil/3, K) :- K := 101.
internal(set_user_id/2, K) :- K := 102.
internal(_, K) :- otherwise | K := none.

special(','/2, F) :- F := yes.
special(is/2, F) :- F := yes.
special(':='/2, F) :- F := yes.
special(_, F) :- otherwise | F := user.
