% Name/topology server node
%
% state = {nodes, topologies, requests}
% nodes = {NODES, TUNNELS, COUNT}
% topologies = {MINCOUNT, RING, TORUS}
% requests =  {ALL, REQUESTS}
%
% NODES is just a list of addresses.
% RING/TORUS look like this:
%
%   [[NODE1, [PEER1|NODE2], ...], ...]
%
% REQUESTS is a list of {PEER, SRC, RNODE} tuples holding
% requests for peers that could not be resolved, yet.
%
% COUNT gives the number of nodes registered, MINCOUNT is
% the number of nodes which must be registered before
% peer requests are processed.
%
% ALL is instantiated to NODES, once C >= MC.
%
% A single numeric argument on the command-line (following "--")
% can be used to set MINCOUNT.

-exports(['$start'/0]).

'$start' :- 
    '$events'(ES), 
    command_line(ARGS),
    parse_args(ARGS, MINCOUNT),
    initial_state(MINCOUNT, ST),
    loop(ES, ST).

parse_args([], MC) :- MC := 0.
parse_args(['--', MCOUNT], MC) :-
    string_to_integer(MCOUNT, MC).
parse_args([X|R], MC) :- X =\= '--' | parse_args(R, MC).
parse_args(_, _) :-
    otherwise | 
    error(usage('strandvm OPTIONS ... ns [-- MINCOUNT]')).    

initial_state(MC, ST) :- ST := {{[], [], 0}, {MC, [], []}, {_, []}}.

% handle ns-specific events
loop([], _) :- '$log'('event stream closed').
loop([register(NODE)|ES], {{ALL, TUNS, C}, TOPS, RS}) :-
    register_node(ALL, NODE, local, ALL2),
    C1 is C + 1,
    propagate_node(TUNS, register_remote(NODE)),
    do_pending_requests({{ALL2, TUNS, C1}, TOPS, RS}, ES).
loop([register_remote(NODE)|ES], {{ALL, TUNS, C}, TOPS, RS}) :-
    % not propagated
    register_node(ALL, NODE, remote, ALL2),
    C1 is C + 1,
    do_pending_requests({{ALL2, TUNS, C1}, TOPS, RS}, ES).
loop([register_tunnel(NODE)|ES], {{ALL, TUNS, C}, TOPS, RS}) :-
    register_node(TUNS, NODE, tunnel, TUNS2),
    RNS is NODE * 256,
    propagate_all_nodes(ALL, RNS),
    loop(ES, {{ALL, TUNS2, C}, TOPS, RS}).
loop([topology(TYPE, TOP)|ES], ST) :-
    update_topology(TYPE, TOP, ST, ST2),
    loop(ES, ST2).
loop([lookup(PEER, RNODE, SRC)|ES], ST) :-
    counts(ST, C, MC),
    perform_lookup(PEER, RNODE, SRC, ES, C, MC, ST).
loop([E|ES], ST) :- handle(E, ST), loop(ES, ST).

perform_lookup(PEER, RNODE, SRC, ES, C, MC, {NS, TOPS, {ALL,
    RS}}) :-
    C < MC |
    '$log'(lookup_postponed(PEER, SRC, C, MC)),
    add_peer_request(RS, PEER, SRC, RNODE, RS2),
    loop(ES, {NS, TOPS, {ALL, RS2}}).
perform_lookup(PEER, RNODE, SRC, ES, C, MC, ST) :-
    C >= MC |
    '$log'(lookup(PEER, SRC)),
    lookup_peer(PEER, SRC, RNODE, ST, ST2),
    loop(ES, ST2).

% process pending events for resolvable peer
do_pending_requests(ST, ES) :-
    counts(ST, C, MC),
    do_pending_requests1(C, MC, ST, ES).

do_pending_requests1(C, MC, ST, ES) :-
    C < MC |
    loop(ES, ST).
do_pending_requests1(C, MC, {NS, TOPS, {ALL, RS}}, ES) :-
    C >= MC |
    set_all(NS, ALL),
    do_pending_requests2(RS, {NS, TOPS, {ALL, []}}, ES).

do_pending_requests2([], ST, ES) :- loop(ES, ST).
do_pending_requests2([{PEER, NODE, R}|RS], ST, ES) :-
    lookup_peer(PEER, NODE, R, ST, ST2),
    do_pending_requests2(RS, ST2, ES).

set_all({NODES, _, _}, ALL) :- 
    unknown(ALL) | ALL := NODES.
set_all(_, _) :- otherwise | true.

% internal events
handle(send_read(ID, PID), _) :-
    '$node'(SELF), 
    send_with(SELF, PID, read(ID, PID, SELF)).
handle(send_value(ID, PID, X), _) :-
    '$send'(PID, value(ID, PID, X)).
handle(send_addref(PID, ID), _) :-
    '$send'(PID, addref(ID)).

% external events
handle(read(ID, SELF, PID), _) :-
    '$get_exposed'(ID, X), 
    send_value(X, PID, ID, SELF).  % suspends on X
handle(value(ID, OWNER, X), _) :- 
    '$assign_remote'(ID, OWNER, X).
handle(addref(ID), _) :- '$add_reference'(ID).
handle(nodes(NODES), {_, _, {ALL, _}}) :- 
    NODES := ALL.
handle(terminate, ST) :-
    all_nodes(ST, NODES),
    broadcast(NODES, terminate, DONE),
    tunnels(ST, TUNS),
    terminate_remote_ns(TUNS, DONE, DONE2),
    '$node'(SELF),
    LMAC is SELF / 256,
    terminate_tunnels(TUNS, LMAC, DONE2).
handle(terminate_ns, _) :- halt(0).
handle(E, _) :- 
    otherwise | '$log'(unkown_event(E)).

terminate_tunnels([], _, []) :- halt(0).
terminate_tunnels([T|TUNS], LMAC, []) :-
    TA is LMAC * 256 + T,
    '$send'(TA, terminate, DONE),
    terminate_tunnels(TUNS, LMAC, DONE).

terminate_remote_ns([], [], DONE) :- DONE := [].
terminate_remote_ns([T|TUNS], [], DONE) :-
    RNS is T * 256,    
    '$send'(RNS, terminate_ns, SENT),
    terminate_remote_ns(TUNS, SENT, DONE).

send_with(X, DEST, MSG) :- 
    data(X) | '$send'(DEST, MSG).

send_value(X, PID, ID, SELF) :-
    data(X) |
    '$drop_reference'(ID),
    '$send'(PID, value(ID, SELF, X)).

lookup_peer(fwd, NODE, R, ST, ST2) :-
    lookup_ring(fwd, NODE, R, ST, ST2).
lookup_peer(bwd, NODE, R, ST, ST2) :-
    lookup_ring(bwd, NODE, R, ST, ST2).
lookup_peer(north, NODE, R, ST, ST2) :-
    lookup_torus(north, NODE, R, ST, ST2).
lookup_peer(east, NODE, R, ST, ST2) :-
    lookup_torus(east, NODE, R, ST, ST2).
lookup_peer(south, NODE, R, ST, ST2) :-
    lookup_torus(south, NODE, R, ST, ST2).
lookup_peer(west, NODE, R, ST, ST2) :-
    lookup_torus(west, NODE, R, ST, ST2).
lookup_peer(PEER, _, _, ST, ST2) :-
    otherwise |
    '$log'(unknown_peer(PEER)),
    ST2 := ST.

lookup_ring(PEER, NODE, R, {{ALL, TUNS, C}, {MC, [], TT},
    RS}, ST) :-
    topology:ring(ALL, RT),
    propagate_topology(TUNS, ring, RT),
    lookup_peer2(RT, PEER, NODE, R, {{ALL, TUNS, C},
        {MC, RT, TT}, RS}, ST).
lookup_ring(PEER, NODE, R, ST, ST2) :-
    otherwise |
    topology(ring, ST, RING),
    lookup_peer2(RING, PEER, NODE, R, ST, ST2).

lookup_torus(PEER, NODE, R, {{ALL, TUNS, C},
    {MC, RT, []}, RS}, ST) :-
    topology:torus(ALL, TT),
    propagate_topology(TUNS, torus, TT),
    lookup_peer2(TT, PEER, NODE, R, {{ALL, TUNS, C},
        {MC, RT, TT}, RS}, ST).
lookup_torus(PEER, NODE, R, ST, ST2) :-
    otherwise |
    topology(torus, ST, TORUS),
    lookup_peer2(TORUS, PEER, NODE, R, ST, ST2).

lookup_peer2([], PEER, NODE, R, {NS, TOPS, {ALL, RS}}, ST) :-
    add_peer_request(RS, PEER, NODE, R, RS2),
    ST := {NS, TOPS, {ALL, RS2}},
    '$log'(peer_request_pending(PEER, NODE)).    
lookup_peer2([[NODE|PEERS]|_], PEER, NODE, R, ST, ST2) :-
    lookup_peer_by_name(PEERS, PEER, NODE, R, no, ST, ST2).
lookup_peer2([[NODE1|_]|ALL], PEER, NODE, R, ST, ST2) :-
    NODE1 =\= NODE |
    lookup_peer2(ALL, PEER, NODE, R, ST, ST2).

lookup_peer_by_name([], PEER, NODE, R, yes, {NS, TOPS, {ALL, 
    RS}}, ST) :-
    add_peer_request(RS, PEER, NODE, R, RS2),
    ST := {NS, TOPS, {ALL, RS2}},
    '$log'(peer_request_pending(PEER, NODE)).    
lookup_peer_by_name([], PEER, NODE, _, no, ST, ST2) :-
    ST2 := ST,
    '$log'(no_peer_for_node(PEER, NODE)).    
lookup_peer_by_name([[PEER|NODE2]|_], PEER, NODE, R, _, ST, ST2) :-
    R := NODE2,
    ST2 := ST,
    '$log'(peer_found(PEER, NODE2, NODE)),
    '$send'(NODE, peer(PEER, NODE2)).
lookup_peer_by_name([[PEER1|_]|PEERS], PEER, NODE, R, M, 
    ST, ST2) :-
    PEER1 =\= PEER |
    lookup_peer_by_name(PEERS, PEER, NODE, R, M, ST, ST2).

% accessors
topology(ring, {_, {_, RING, _}, _}, T) :- T := RING.

topology(torus, {_, {_, _, TORUS}, _}, T) :- T := TORUS.

all_nodes({{ALL, _, _}, _, _}, R) :- R := ALL.

tunnels({{_, TUNS, _}, _, _}, R) :- R := TUNS.

counts({{_, _, C}, {MC, _, _}, _}, CR, MCR) :-
    CR := C, MCR := MC.

update_topology(ring, RT, {NS, {MC, _, TT}, RS}, ST2) :-
    ST2 := {NS, {MC, RT, TT}, RS}.
update_topology(torus, TT, {NS, {MC, RT, _}, RS}, ST2) :-
    ST2 := {NS, {MC, RT, TT}, RS}.

register_node([], NODE, TYPE, ALL) :-
    '$log'(register(TYPE, NODE)),
    ALL := [NODE].
register_node([NODE|ALL], NODE, TYPE, ALL2) :-
    '$log'(re_register(TYPE, NODE)),
    ALL2 := [NODE|ALL].
register_node([NODE1|ALL], NODE, TYPE, ALL2) :-
    ALL2 := [NODE1|ALL3],
    register_node(ALL, NODE, TYPE, ALL3).

propagate_node([], _).
propagate_node([T|TUNS], MSG) :-
    RNS is T * 256,
    '$send'(RNS, MSG),
    propagate_node(TUNS, MSG).

propagate_all_nodes([], _).
propagate_all_nodes([N|ALL], RNS) :-
    '$send'(RNS, register_remote(N)),
    propagate_all_nodes(ALL, RNS).

propagate_topology([], _, _).
propagate_topology([T|TUNS], TYPE, TN) :-
    RNS is T * 256,
    send_with(TN, RNS, topology(TYPE, TN)),
    propagate_topology(TUNS, TYPE, TN).

broadcast([], _, DONE) :- DONE := [].
broadcast([N|ALL], MSG, DONE) :-
    '$send'(N, MSG),
    broadcast(ALL, MSG, DONE).

add_peer_request([], PEER, NODE, R, RS) :-
    RS := [{PEER, NODE, R}].
add_peer_request([{P, N, R}|RS], P, N, _, RS2) :-
    RS2 := [{P, N, R}|RS].
add_peer_request([{P1, N1, R1}|RS], P, N, R, RS2) :-
    otherwise |
    RS2 := [{P1, N1, R1}|RS3],
    add_peer_request(RS, P, N, R, RS3).
