% system glue library for inter-node communication
%
% init
%   Starts event handling (done automatically when loaded from 
%   command line).
%
% handle(EVENT)
%   Handle single standard-protocol event.
%
% self(ADDR^)
%   Assign own address to ADDR.
%
% nodes(ALL^)
%   returns a list of all node addresses in the namespace.
%
% terminate
%   terminate all registered nodes, and stop name server.

-exports(['$start'/0, init/0, handle/1, nodes/1, terminate/0, self/1]).

'$start' :- init.

init :-
    '$events'(ES), 
    '$node'(SELF),
    send_with(SELF, 0, register(SELF)),
    loop(ES).

loop([]) :- '$log'('event stream closed').
loop([E|ES]) :- handle(E), loop(ES).

% internal events
handle(send_read(ID, PID)) :-
    '$node'(SELF), 
    send_with(SELF, PID, read(ID, PID, SELF)).
handle(send_value(ID, PID, X)) :-
    '$send'(PID, value(ID, PID, X)).
handle(send_get_module(NID, MID, MOD)) :-
    '$send'(NID, get_module(MID, MOD)).
handle(send_call(MID, IDX, ARGS, PEER)) :-
    '$resolve_peer'(PEER, PID),
    '$node'(SELF),
    send_with(SELF, PID, call(MID, IDX, SELF, ARGS)).
handle(send_lookup(NAME, PEER)) :-
    '$node'(SELF),
    send_with(SELF, 0, lookup(NAME, PEER, SELF)).
handle(send_addref(PID, ID)) :-
    '$send'(PID, addref(ID)).
% external events
handle(read(ID, SELF, PID)) :-
    '$get_exposed'(ID, X), 
    send_value(X, PID, ID, SELF).  % suspends on X
handle(value(ID, OWNER, X)) :- 
    '$assign_remote'(ID, OWNER, X).
handle(call(MID, IDX, NID, ARGS)) :-
    '$retrieve_module'(MID, NID, MOD),
    '$call'(MOD, IDX, ARGS).
handle(get_module(MID, MOD)) :-
    '$load_module'(MID, MOD).
handle(peer(PEER, NODE)) :-
    '$register_peer'(PEER, NODE).
handle(terminate) :- halt(0).
handle(exit(REASON)) :-
    '$node'(SELF, PID),
    fmt:format(2, '[~q:~q] terminated - ~q~n', [PID, SELF, REASON],
        DONE),
    done(DONE, 1).
handle(addref(ID)) :-
    '$add_reference'(ID).
handle(E) :-
    otherwise |
    error(unkown_event(E)).

done([], CODE) :- halt(CODE).

send_with(SELF, DEST, MSG) :- 
    data(SELF) | '$send'(DEST, MSG).

send_value(X, PID, ID, SELF) :-
    data(X) |
    '$drop_reference'(ID),
    '$send'(PID, value(ID, SELF, X)).

% exported interface

nodes(ALL) :- '$send'(0, nodes(ALL)).
terminate :- '$send'(0, terminate).
self(SELF) :- '$node'(SELF).
